import asyncio
import os
import uuid

import pytest
from binascii import hexlify

import aiormq
from aiormq.auth import AuthBase, PlainAuth
from .conftest import skip_when_quick_test


pytestmark = pytest.mark.asyncio

CERT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), 'certs'))


pytest.mark.parametrize(
    "connection",
)


async def test_simple(amqp_connection: aiormq.Connection):
    assert amqp_connection.reader is not None
    assert amqp_connection.writer is not None

    channel1 = await amqp_connection.channel()
    await channel1.basic_qos(prefetch_count=1)

    assert channel1.number

    channel2 = await amqp_connection.channel(11)
    assert channel2.number

    await channel1.close()
    await channel2.close()

    channel = await amqp_connection.channel()

    queue = asyncio.Queue()

    deaclare_ok = await channel.queue_declare(auto_delete=True)
    consume_ok = await channel.basic_consume(deaclare_ok.queue, queue.put)
    await channel.basic_publish(b'foo', routing_key=deaclare_ok.queue)

    message = await queue.get()     # type: DeliveredMessage
    assert message.body == b'foo'

    with pytest.raises(aiormq.exceptions.DeliveryError) as e:
        await channel.basic_publish(
            b'bar', routing_key=deaclare_ok.queue + 'foo',
            mandatory=True,
        )

    message = e.value.message

    assert message.delivery.routing_key == deaclare_ok.queue + 'foo'
    assert message.body == b'bar'

    cancel_ok = await channel.basic_cancel(consume_ok.consumer_tag)
    assert cancel_ok.consumer_tag == consume_ok.consumer_tag
    await channel.queue_delete(deaclare_ok.queue)

    deaclare_ok = await channel.queue_declare(auto_delete=True)
    await channel.basic_publish(b'foo bar', routing_key=deaclare_ok.queue)

    message = await channel.basic_get(deaclare_ok.queue, no_ack=True)
    assert message.body == b'foo bar'

    await amqp_connection.close()

    with pytest.raises(RuntimeError):
        await channel.basic_get(deaclare_ok.queue)

    with pytest.raises(RuntimeError):
        await amqp_connection.channel()

    assert amqp_connection.reader is None
    assert amqp_connection.writer is None


async def test_bad_channel(amqp_connection: aiormq.Connection):
    with pytest.raises(ValueError):
        await amqp_connection.channel(65536)

    with pytest.raises(ValueError):
        await amqp_connection.channel(-1)

    channel = await amqp_connection.channel(65535)

    with pytest.raises(aiormq.exceptions.ChannelNotFoundEntity):
        await channel.queue_bind(
            uuid.uuid4().hex,
            uuid.uuid4().hex
        )

    await asyncio.sleep(0.1)

    assert amqp_connection.channels[channel.number] is None

    with pytest.raises(aiormq.exceptions.ConnectionChannelError):
        await amqp_connection.channel(65535)


async def test_properties(amqp_connection):
    assert amqp_connection.connection_tune.channel_max > 0
    assert amqp_connection.connection_tune.heartbeat > 1
    assert amqp_connection.connection_tune.frame_max > 1024
    await amqp_connection.close()


async def test_open(amqp_connection):
    with pytest.raises(RuntimeError):
        await amqp_connection.connect()

    channel = await amqp_connection.channel()
    await channel.close()
    await amqp_connection.close()


async def test_auth_base(amqp_connection):
    with pytest.raises(NotImplementedError):
        AuthBase(amqp_connection).marshal()


async def test_auth_plain(amqp_connection):
    auth = PlainAuth(amqp_connection).marshal()

    assert auth == PlainAuth(amqp_connection).marshal()

    auth_parts = auth.split(b"\x00")
    assert auth_parts == [b'', b'guest', b'guest']

    connection = aiormq.Connection(
        amqp_connection.url.with_user('foo').with_password('bar')
    )

    auth = PlainAuth(connection).marshal()

    auth_parts = auth.split(b"\x00")
    assert auth_parts == [b'', b'foo', b'bar']

    auth = PlainAuth(connection)
    auth.value = b'boo'

    assert auth.marshal() == b'boo'


async def test_channel_closed(amqp_connection):

    channel = await amqp_connection.channel()

    with pytest.raises(aiormq.exceptions.ChannelNotFoundEntity):
        await channel.basic_consume("foo", lambda x: None)

    await amqp_connection.close()


async def test_bad_credentials(amqp_connection):
    connection = aiormq.Connection(
        amqp_connection.url.with_password(uuid.uuid4().hex),
    )

    with pytest.raises(aiormq.exceptions.ProbableAuthenticationError):
        await connection.connect()


async def test_non_publisher_confirms(amqp_connection):
    amqp_connection.server_capabilities['publisher_confirms'] = False

    with pytest.raises(ValueError):
        await amqp_connection.channel(publisher_confirms=True)

    await amqp_connection.channel(publisher_confirms=False)


@skip_when_quick_test
async def test_no_free_channels(amqp_connection):
    await asyncio.wait([
        amqp_connection.channel(n + 1)
        for n in range(amqp_connection.connection_tune.channel_max)
    ])

    with pytest.raises(aiormq.exceptions.ConnectionNotAllowed):
        await amqp_connection.channel()


async def test_huge_message(amqp_connection: aiormq.Connection):
    conn = amqp_connection      # type: aiormq.Connection
    body = os.urandom(int(conn.connection_tune.frame_max * 2.5))
    channel = await conn.channel()       # type: aiormq.Channel

    queue = asyncio.Queue()

    deaclare_ok = await channel.queue_declare(auto_delete=True)
    await channel.basic_consume(deaclare_ok.queue, queue.put)
    await channel.basic_publish(body, routing_key=deaclare_ok.queue)

    message = await queue.get()  # type: DeliveredMessage
    assert message.body == body


async def test_return_message(amqp_connection: aiormq.Connection):
    conn = amqp_connection      # type: aiormq.Connection
    body = os.urandom(512)
    routing_key = hexlify(os.urandom(16)).decode()

    channel = await conn.channel(
        on_return_raises=False
    )   # type: aiormq.Channel

    result = await channel.basic_publish(
        body,
        routing_key=routing_key,
        mandatory=True
    )

    assert result.delivery.name == 'Basic.Return'
    assert result.delivery.routing_key == routing_key


async def test_cancel_on_queue_deleted(amqp_connection, event_loop):
    conn = amqp_connection  # type: aiormq.Connection
    channel = await conn.channel()  # type: aiormq.Channel
    deaclare_ok = await channel.queue_declare(auto_delete=True)
    consume_ok = await channel.basic_consume(deaclare_ok.queue, print)

    assert consume_ok.consumer_tag in channel.consumers

    with pytest.raises(aiormq.DuplicateConsumerTag):
        await channel.basic_consume(
            deaclare_ok.queue, print,
            consumer_tag=consume_ok.consumer_tag
        )

    await channel.queue_delete(deaclare_ok.queue)

    await asyncio.sleep(0.1, loop=event_loop)

    assert consume_ok.consumer_tag not in channel.consumers



import asyncio
import os

import pytest
from async_generator import yield_, async_generator
from yarl import URL

from aiormq import Connection


@pytest.fixture
def event_loop():
    asyncio.get_event_loop().close()
    loop = asyncio.new_event_loop()     # type: asyncio.AbstractEventLoop
    loop.set_debug(True)
    asyncio.set_event_loop(loop)
    try:
        yield loop
    finally:
        loop.close()


def cert_path(*args):
    return os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'certs', *args
    )


amqp_urls = [
    URL(os.getenv('AMQP_URL', 'amqp://guest:guest@localhost/')),
    URL(os.getenv(
        'AMQP_URL', 'amqp://guest:guest@localhost/'
    )).with_scheme('amqps').with_query({
        "cafile": cert_path('ca.pem'),
        "no_verify_ssl": 1,
    }),
    URL(os.getenv(
        'AMQP_URL', 'amqp://guest:guest@localhost/'
    )).with_scheme('amqps').with_query({
        "cafile": cert_path('ca.pem'),
        "keyfile": cert_path('client.key'),
        "certfile": cert_path('client.pem'),
        "no_verify_ssl": 1,
    })
]


@pytest.fixture(params=amqp_urls)
@async_generator
async def amqp_connection(request, event_loop):
    connection = Connection(request.param, loop=event_loop)

    await connection.connect()

    try:
        await yield_(connection)
    finally:
        await connection.close()


channel_params = [
    dict(channel_number=None, frame_buffer=10, publisher_confirms=True),
    dict(channel_number=None, frame_buffer=1, publisher_confirms=True),
    dict(channel_number=None, frame_buffer=10, publisher_confirms=False),
    dict(channel_number=None, frame_buffer=1, publisher_confirms=False),
]


@pytest.fixture(params=channel_params)
@async_generator
async def amqp_channel(request, amqp_connection):
    try:
        await yield_(await amqp_connection.channel(**request.param))
    finally:
        await amqp_connection.close()


skip_when_quick_test = pytest.mark.skipif(
    os.getenv("TEST_QUICK") is not None,
    reason='quick test'
)




class AMQPError(Exception):
    message = 'An unspecified AMQP error has occurred'

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.message % self.args)


class AMQPConnectionError(AMQPError):
    message = 'Connection can not be opened'


class IncompatibleProtocolError(AMQPConnectionError):
    message = 'The protocol returned by the server is not supported'


class AuthenticationError(AMQPConnectionError):
    message = (
        'Server and client could not negotiate use of the '
        'authentication mechanisms. Server supports only %r, '
        'but client supports only %r.'
    )


class ProbableAuthenticationError(AMQPConnectionError):
    message = (
        'Client was disconnected at a connection stage indicating a '
        'probable authentication error: %s'
    )


class ConnectionClosed(AMQPConnectionError):
    message = 'The AMQP connection was closed (%s) %s'


class ConnectionSyntaxError(ConnectionClosed):
    message = ('The sender sent a frame that contained illegal values for '
               'one or more fields. This strongly implies a programming error '
               'in the sending peer: %r')


class ConnectionFrameError(ConnectionClosed):
    message = ('The sender sent a malformed frame that the recipient could '
               'not decode. This strongly implies a programming error '
               'in the sending peer: %r')


class ConnectionCommandInvalid(ConnectionClosed):
    message = ('The client sent an invalid sequence of frames, attempting to '
               'perform an operation that was considered invalid by the server.'
               ' This usually implies a programming error in the client: %r')


class ConnectionChannelError(ConnectionClosed):
    message = ('The client attempted to work with a channel that had not been '
               'correctly opened. This most likely indicates a fault in the '
               'client layer: %r')


class ConnectionUnexpectedFrame(ConnectionClosed):
    message = ("The peer sent a frame that was not expected, usually in the "
               "context of a content header and body. This strongly indicates "
               "a fault in the peer's content processing: %r")


class ConnectionResourceError(ConnectionClosed):
    message = ("The server could not complete the method because it lacked "
               "sufficient resources. This may be due to the client creating "
               "too many of some type of entity: %r")


class ConnectionNotAllowed(ConnectionClosed):
    message = ("The client tried to work with some entity in a manner that is "
               "prohibited by the server, due to security settings or by "
               "some other criteria: %r")


class ConnectionNotImplemented(ConnectionClosed):
    message = ("The client tried to use functionality that is "
               "not implemented in the server: %r")


class ConnectionInternalError(ConnectionClosed):
    message = (" The server could not complete the method because of an "
               "internal error. The server may require intervention by an "
               "operator in order to resume normal operations: %r")


class AMQPChannelError(AMQPError):
    message = 'An unspecified AMQP channel error has occurred'


class ChannelAccessRefused(AMQPChannelError):
    message = ('The client attempted to work with a server entity to '
               'which it has no access due to security settings: %r')


class ChannelNotFoundEntity(AMQPChannelError):
    message = ('The client attempted to work with a server '
               'entity that does not exist: %r')


class ChannelLockedResource(AMQPChannelError):
    message = ('The client attempted to work with a server entity to '
               'which it has no access because another client is working '
               'with it: %r')


class ChannelPreconditionFailed(AMQPChannelError):
    message = ('The client requested a method that was not allowed because '
               'some precondition failed: %r')


class ChannelClosed(AMQPChannelError):
    message = 'The channel was closed (%s) %s'


class DuplicateConsumerTag(AMQPChannelError):
    message = 'The consumer tag specified already exists for this channel: %s'


class ProtocolSyntaxError(AMQPError):
    message = 'An unspecified protocol syntax error occurred'


class InvalidFrameError(ProtocolSyntaxError):
    message = 'Invalid frame received: %r'


class MethodNotImplemented(AMQPError):
    pass


class AMQPException(Exception):
    pass


class DeliveryError(AMQPException):
    __slots__ = 'message', 'frame'

    def __init__(self, message, frame):
        self.message = message
        self.frame = frame



import asyncio
import logging
import platform
import ssl
import typing
from contextlib import suppress

import pamqp.frame
from pamqp import ProtocolHeader
from pamqp import specification as spec
from pamqp.heartbeat import Heartbeat
from yarl import URL

from aiormq.channel import Channel
from . import exceptions as exc
from .auth import AuthMechanism
from .base import Base, task
from .tools import censor_url
from .types import (
    ArgumentsType, SSLCerts,
    URLorStr)
from .version import __version__

log = logging.getLogger(__name__)


try:
    from yarl import DEFAULT_PORTS

    DEFAULT_PORTS['amqp'] = 5672
    DEFAULT_PORTS['amqps'] = 5671
except ImportError:
    pass


PRODUCT = 'aiormq'
PLATFORM = '{} {} ({} build {})'.format(
    platform.python_implementation(),
    platform.python_version(),
    *platform.python_build()
)


class Connection(Base):
    FRAME_BUFFER = 10

    def __init__(self, url: URLorStr, *, parent=None,
                 loop: asyncio.get_event_loop() = None):

        super().__init__(
            loop=loop or asyncio.get_event_loop(),
            parent=parent
        )

        self.url = URL(url)
        self.vhost = self.url.path.strip("/") or "/"
        self.reader = None  # type: asyncio.StreamReader
        self.writer = None  # type: asyncio.StreamWriter
        self.ssl_certs = SSLCerts(
            ca=self.url.query.get('cafile'),
            key=self.url.query.get('keyfile'),
            cert=self.url.query.get('certfile'),
            verify=self.url.query.get('no_verify_ssl', '0') == '0'
        )

        self.started = False
        self.__lock = asyncio.Lock(loop=self.loop)

        self.channels = {}  # type: typing.Dict[int, typing.Optional[Channel]]

        self.server_properties = None   # type: spec.Connection.OpenOk
        self.connection_tune = None  # type: spec.Connection.TuneOk

        self.last_channel = 0

    @property
    def lock(self):
        if self.is_closed:
            raise RuntimeError('%r closed' % self)

        return self.__lock

    @property
    def is_opened(self):
        return self.writer is not None and not self.is_closed

    def __str__(self):
        return str(censor_url(self.url))

    def _get_ssl_context(self):
        context = ssl.create_default_context(
            (
                ssl.Purpose.SERVER_AUTH
                if self.ssl_certs.key
                else ssl.Purpose.CLIENT_AUTH
            ),
            capath=self.ssl_certs.ca,
        )

        if self.ssl_certs.key:
            context.load_cert_chain(
                self.ssl_certs.cert,
                self.ssl_certs.key,
            )

        if not self.ssl_certs.verify:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        return context

    @staticmethod
    def _client_capabilities():
        return {
            'platform': PLATFORM,
            'version': __version__,
            'product': PRODUCT,
            'capabilities': {
                'authentication_failure_close': True,
                'basic.nack': True,
                'connection.blocked': False,
                'consumer_cancel_notify': True,
                'publisher_confirms': True
            },
            'information': 'See https://github.com/mosquito/aiormq/',
        }

    @staticmethod
    def _credentials_class(start_frame: spec.Connection.Start):
        for mechanism in start_frame.mechanisms.decode().split():
            with suppress(KeyError):
                return AuthMechanism[mechanism]

        raise exc.AuthenticationError(
            start_frame.mechanisms,
            [m.name for m in AuthMechanism]
        )

    async def __rpc(self, request: spec.Frame, wait_response=True):
        self.writer.write(pamqp.frame.marshal(request, 0))

        if not wait_response:
            return

        _, _, frame = await self.__receive_frame()

        if request.synchronous and frame.name not in request.valid_responses:
            raise spec.AMQPInternalError(frame, frame)
        elif isinstance(frame, spec.Connection.Close):
            if frame.reply_code == 403:
                raise exc.ProbableAuthenticationError(frame.reply_text)

            raise exc.ConnectionClosed(frame.reply_code, frame.reply_text)
        return frame

    @task
    async def connect(self):
        if self.writer is not None:
            raise RuntimeError("Already connected")

        ssl_context = None

        if self.url.scheme == 'amqps':
            ssl_context = await self.loop.run_in_executor(
                None, self._get_ssl_context
            )

        try:
            self.reader, self.writer = await asyncio.open_connection(
                self.url.host, self.url.port, ssl=ssl_context,
                loop=self.loop
            )
        except OSError as e:
            raise ConnectionError(*e.args) from e

        try:
            protocol_header = ProtocolHeader()
            self.writer.write(protocol_header.marshal())

            res = await self.__receive_frame()
            _, _, frame = res   # type: spec.Connection.Start
        except EOFError as e:
            raise exc.IncompatibleProtocolError(*e.args) from e

        credentials = self._credentials_class(frame)

        self.server_properties = frame.server_properties

        self.connection_tune = await self.__rpc(spec.Connection.StartOk(
            client_properties=self._client_capabilities(),
            mechanism=credentials.name,
            response=credentials.value(self).marshal()
        ))      # type: spec.Connection.Tune

        await self.__rpc(spec.Connection.TuneOk(
            channel_max=self.connection_tune.channel_max,
            frame_max=self.connection_tune.frame_max,
            heartbeat=self.connection_tune.heartbeat,
        ), wait_response=False)

        await self.__rpc(spec.Connection.Open(virtual_host=self.vhost))

        # noinspection PyAsyncCall
        self.create_task(self.__reader())

        return True

    @task
    async def __receive_frame(self) -> typing.Tuple[int, int, spec.Frame]:
        async with self.lock:
            frame_header = await self.reader.readexactly(1)

            if frame_header == b'\0x00':
                raise spec.AMQPFrameError(await self.reader.read())

            frame_header += await self.reader.readexactly(6)

            if not self.started and frame_header.startswith(b'AMQP'):
                raise spec.AMQPSyntaxError
            else:
                self.started = True

            # noinspection PyProtectedMember
            frame_type, _, frame_length = pamqp.frame._frame_parts(
                frame_header
            )

            frame_payload = await self.reader.readexactly(
                frame_length + 1
            )

            return pamqp.frame.unmarshal(frame_header + frame_payload)

    @staticmethod
    def __exception_by_code(frame: spec.Connection.Close):
        if frame.reply_code == 501:
            return exc.ConnectionFrameError(frame.reply_text)
        elif frame.reply_code == 502:
            return exc.ConnectionSyntaxError(frame.reply_text)
        elif frame.reply_code == 503:
            return exc.ConnectionCommandInvalid(frame.reply_text)
        elif frame.reply_code == 504:
            return exc.ConnectionChannelError(frame.reply_text)
        elif frame.reply_code == 505:
            return exc.ConnectionUnexpectedFrame(frame.reply_text)
        elif frame.reply_code == 506:
            return exc.ConnectionResourceError(frame.reply_text)
        elif frame.reply_code == 530:
            return exc.ConnectionNotAllowed(frame.reply_text)
        elif frame.reply_code == 540:
            return exc.ConnectionNotImplemented(frame.reply_text)
        elif frame.reply_code == 541:
            return exc.ConnectionInternalError(frame.reply_text)
        else:
            return exc.ConnectionClosed(frame.reply_code, frame.reply_text)

    async def __reader(self):
        while self.writer:
            try:
                weight, channel, frame = await self.__receive_frame()

                if channel == 0:
                    if isinstance(frame, Heartbeat):
                        self.writer.write(
                            pamqp.frame.marshal(Heartbeat(), channel)
                        )
                        continue
                    elif isinstance(frame, spec.Connection.Close):
                        return await self.close(self.__exception_by_code(frame))

                    log.error('Unexpected frame %r', frame)
                    continue

                queue = self.channels[channel].frames
                await queue.put((weight, frame))
            except asyncio.CancelledError:
                return

    async def _on_close(self, exc=exc.ConnectionClosed(0, 'normal closed')):
        writer = self.writer
        self.reader = None
        self.writer = None

        # noinspection PyShadowingNames
        writer.close()
        return await writer.wait_closed()

    @property
    def server_capabilities(self) -> ArgumentsType:
        return self.server_properties['capabilities']

    @property
    def basic_nack(self) -> bool:
        return self.server_capabilities.get('basic.nack')

    @property
    def consumer_cancel_notify(self) -> bool:
        return self.server_capabilities.get('consumer_cancel_notify')

    @property
    def exchange_exchange_bindings(self) -> bool:
        return self.server_capabilities.get('exchange_exchange_bindings')

    @property
    def publisher_confirms(self):
        return self.server_capabilities.get('publisher_confirms')

    async def channel(self, channel_number: int = None,
                      publisher_confirms=True,
                      frame_buffer=FRAME_BUFFER, **kwargs) -> Channel:

        if self.is_closed:
            raise RuntimeError('%r closed' % self)

        if not self.publisher_confirms and publisher_confirms:
            raise ValueError("Server doesn't support publisher_confirms")

        if channel_number is None:
            self.last_channel += 1

            while self.last_channel in self.channels.keys():
                self.last_channel += 1

            channel_number = self.last_channel

        if channel_number < 0 or channel_number > 65535:
            raise ValueError('Channel number too large')

        self.last_channel = min(self.last_channel, channel_number)

        channel = Channel(
            self, channel_number, frame_buffer=frame_buffer,
            publisher_confirms=publisher_confirms, **kwargs
        )

        self.channels[channel_number] = channel

        try:
            await channel.open()
        except Exception:
            self.channels.pop(channel_number, None)
            raise

        def on_close(*_):
            self.channels[channel_number] = None

        channel.closing.add_done_callback(on_close)

        return channel

    async def __aenter__(self):
        await self.connect()


async def connect(url, *args, **kwargs) -> Connection:
    connection = Connection(url, *args, **kwargs)
    await connection.connect()

    return connection



import typing

from pamqp import specification as spec, ContentHeader
from yarl import URL


ChannelRType = typing.Tuple[int, spec.Channel.OpenOk]
GetResultType = typing.Union[spec.Basic.GetEmpty, spec.Basic.GetOk]
DeliveredMessage = typing.NamedTuple(
    'DeliveredMessage', [
        ('delivery', typing.Union[spec.Basic.Deliver, GetResultType]),
        ('header', ContentHeader),
        ('body', bytes),
        ('channel', "aiormq.Channel")
    ]
)
CallbackCoro = typing.Coroutine[DeliveredMessage, None, typing.Any]
ConsumerCallback = typing.Callable[[], CallbackCoro]
ReturnCallback = typing.Callable[[], CallbackCoro]
ArgumentsType = typing.Dict[str, typing.Union[str, int, bool]]
ConfirmationFrameType = typing.Union[spec.Basic.Ack, spec.Basic.Nack]
SSLCerts = typing.NamedTuple(
    'SSLCerts', [
        ('cert', str),
        ('key', str),
        ('ca', str),
        ('verify', bool),
    ]
)
FrameReceived = typing.NamedTuple(
    'FrameReceived', [
        ('channel', int),
        ('frame', str),
    ]
)


URLorStr = typing.Union[URL, str]



import os

from setuptools import setup, find_packages
from importlib.machinery import SourceFileLoader


module = SourceFileLoader(
    "version", os.path.join("aiormq", "version.py")
).load_module()


setup(
    name='aiormq',
    version=module.__version__,
    packages=find_packages(exclude='tests'),
    license=module.package_license,
    description=module.package_info,
    long_description=open("README.rst").read(),
    url='https://github.com/mosquito/aiormq',
    author=module.__author__,
    author_email=module.team_email,
    install_requires=[
        'pamqp',
        'yarl',
    ],
    keywords=[
        "rabbitmq",
        "asyncio",
        "amqp",
        "amqp 0.9.1",
        "driver",
        "pamqp~=2.1",
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Internet',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Microsoft',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    python_requires=">3.5.*",
    extras_require={
        'develop': [
            'async_generator',
            'coverage!=4.3',
            'coveralls',
            'pylava',
            'pytest',
            'pytest-asyncio',
            'pytest-cov',
            'tox>=2.4',
        ],
    },
)
